import sagemaker
import click
from botocore.exceptions import ClientError
from aws_sagemaker_remote.util.training import training_describe
from .iam import ensure_inference_role
from aws_sagemaker_remote.util.fields import get_field
from aws_sagemaker_remote.ecr.images import ecr_ensure_image
from aws_sagemaker_remote.util.json_read import json_urlparse


def model_delete(name, client):
    client.delete_model(ModelName=name)


def model_describe(name, client, field=None):
    description = client.describe_model(ModelName=name)
    description = get_field(data=description, field=field)
    return description


def model_exists(name, client):
    try:
        model_describe(name=name, client=client)
        return True
    except ClientError:
        return False


def model_create(
    job,
    model_artifact,
    name,
    session: sagemaker.Session,
    inference_image,
    inference_image_path,
    inference_image_accounts,
    role,
    force,
    multimodel=False,
    accelerator_type=None
):
    job = json_urlparse(job)
    name = json_urlparse(name)
    model_artifact = json_urlparse(model_artifact)
    image_uri = ecr_ensure_image(
        path=inference_image_path,
        tag=inference_image,
        accounts=inference_image_accounts.split(","),
        session=session.boto_session
    )
    if (job and model_artifact) or (not (job or model_artifact)):
        raise click.UsageError('Specify one of job_name or model_artifact')
    if model_artifact and not name:
        raise click.UsageError(
            'name is required if job is not provided')
    iam = session.boto_session.client('iam')
    client = session.boto_session.client('sagemaker')
    role = ensure_inference_role(iam=iam, role_name=role)
    if job:
        client = session.boto_session.client('sagemaker')
        model_artifact = training_describe(
            job_name=job,
            field='ModelArtifacts.S3ModelArtifacts',
            client=client
        )
        if not name:
            name = job
        print("Creating model [{}] from job [{}] artifact [{}]".format(
            name, job, model_artifact))
    else:
        if not model_artifact.startswith('s3://'):
            if model_artifact.startswith('/'):
                model_artifact = model_artifact[1:]
            bucket = session.default_bucket()
            model_artifact = 's3://{}/{}'.format(bucket, model_artifact)
        print("Creating model [{}] from artifact [{}]".format(
            name, model_artifact))

    if model_exists(name=name, client=client):
        if force:
            print("Deleting existing model")
            model_delete(name=name, client=client)
        else:
            raise click.UsageError('Specify force if overwriting model')
    model = sagemaker.Model(
        image_uri=image_uri,
        model_data=model_artifact,
        role=role,
        predictor_cls=None,
        env=None,
        name=name,
        # vpc_config=None,
        sagemaker_session=session,
        # enable_network_isolation=False,
        # model_kms_key=None
    )
    container_def = sagemaker.container_def(
        model.image_uri,
        model.model_data,
        model.env,
        container_mode='MultiModel' if multimodel else 'SingleModel')
    """
    client.create_model(
    ModelName='string',
    PrimaryContainer={
        'ContainerHostname': 'string',
        'Image': 'string',
        'ImageConfig': {
            'RepositoryAccessMode': 'Platform'|'Vpc'
        },
        'Mode': 'SingleModel'|'MultiModel',
        'ModelDataUrl': 'string',
        'Environment': {
            'string': 'string'
        },
        'ModelPackageName': 'string'
    },
    """

    # self._ensure_base_name_if_needed(container_def["Image"])
    # self._set_model_name_if_needed()

    enable_network_isolation = model.enable_network_isolation()

    # self._init_sagemaker_session_if_does_not_exist(instance_type)
    session.create_model(
        model.name,
        model.role,
        container_def,
        vpc_config=model.vpc_config,
        enable_network_isolation=enable_network_isolation,
        # tags=tags,
    )



import os
import docker
import base64


class Image(object):
    def __init__(self, name, path, tag, accounts=[]):
        self.name = name
        self.path = path
        self.tag = tag
        self.accounts = accounts


class Images(object):
    INFERENCE = Image(
        'inference',
        os.path.abspath(os.path.join(__file__, '../inference')),
        'aws-sagemaker-remote-inference:latest',
        ['763104351884']
    )

    PROCESSING = Image(
        'processing',
        os.path.abspath(os.path.join(__file__, '../processing')),
        'aws-sagemaker-remote-processing:latest',
        ['763104351884']
    )

    TRAINING = Image(
        'training',
        os.path.abspath(os.path.join(__file__, '../training')),
        'aws-sagemaker-remote-training:latest',
        ['763104351884']
    )

    TRAINING_GPU = Image(
        'training:gpu',
        os.path.abspath(os.path.join(__file__, '../training')),
        'aws-sagemaker-remote-training:gpu',
        ['763104351884']
    )

    ALL = [
        INFERENCE,
        PROCESSING,
        TRAINING,
        TRAINING_GPU
    ]

# Repository management


def ecr_repository_exists(ecr, account, name):
    return ecr_describe_repository(ecr, account, name) is not None


def ecr_describe_repository(ecr, account, name):
    try:
        response = ecr.describe_repositories(
            registryId=account,
            repositoryNames=[
                name,
            ]
        )
    except Exception:
        return None
    repo = response['repositories']
    if len(repo) > 0:
        return repo[0]
    else:
        return None


def ecr_create_repository(ecr, name):
    repo = ecr.create_repository(
        repositoryName=name,
        tags=[
            {
                'Key': 'Source',
                'Value': 'aws-sagemaker-remote'
            },
            # todo: add metadata
        ],
        imageTagMutability='MUTABLE',
        imageScanningConfiguration={
            'scanOnPush': False
        },
        #    encryptionConfiguration={
        #        'encryptionType': 'AES256'|'KMS',
        #        'kmsKey': 'string'
        #    }
    )
    #print("create repo {}".format(repo))
    repo = repo['repository']
    print("Created ECR repository {}".format(name))
    return repo


def ecr_ensure_repo(ecr, account, name, user_account):
    repo = ecr_describe_repository(ecr, account, name)
    if repo is not None:
        return repo
    if account != user_account:
        # todo: check on external accounts
        raise ValueError(
            "Repository {}/{} does not exist and cannot be created because it is in account {} and you are in account {}",
            account, name, account, user_account
        )
    return ecr_create_repository(ecr, name)


def ecr_login(ecr, docker_client, accounts):
    accounts = [account for account in accounts if account]
    token = ecr.get_authorization_token(registryIds=accounts)
    for auth_data in token['authorizationData']:
        username, password = base64.b64decode(
            auth_data['authorizationToken']
        ).decode().split(':')
        registry = auth_data['proxyEndpoint']
        # login =
        docker_client.login(
            username, password, registry=registry, reauth=True)
        print("Logged in to ECR {}".format(registry))
        #print("un/pw: {}:{}, reg: {}, login: {}".format(username, password, registry, login))

# Image operations


def ecr_ensure_image(
    path, tag, accounts, session
):
    ecr = session.client('ecr')  # , region_name='eu-west-1')
    user_account = get_account(session)
    tag_account, tag_repo, tag_tag = parse_image(tag, account=user_account)
    img = get_image(ecr, tag_account, tag_repo, tag_tag)
    if img is not None:
        repo = ecr_ensure_repo(ecr, tag_account, tag_repo, user_account)
        repo_uri = repo['repositoryUri']
        full_uri = "{}:{}".format(repo_uri, tag_tag)
        return full_uri
    else:
        return ecr_build_image(
            path=path,
            tag=tag,
            accounts=accounts,
            session=session,
            pull=False,
            cache=True
        )


def ecr_build_image(
    path, tag, accounts, cache, pull, session
):
    # todo: add region param
    ecr = session.client('ecr')  # , region_name='eu-west-1')
    region_name = session.region_name
    if not region_name:
        raise ValueError("No default region name in your AWS profile")
    user_account = get_account(session)
    tag_account, tag_repo, tag_tag = parse_image(tag, account=user_account)
    repo = ecr_ensure_repo(ecr, tag_account, tag_repo, user_account)
    repo_uri = repo['repositoryUri']
    path = os.path.join(path, tag_tag)

    if not accounts:
        accounts = []
    if tag_account not in accounts:
        accounts = [tag_account] + list(accounts)
    client = docker.from_env()
    ecr_login(ecr, client, accounts)
    image_full = "{}:{}".format(repo_uri, tag_tag)
    print("Building image {} (cache: {}, pull:{})".format(
        image_full,
        cache, pull))
    client.images.build(
        path=path,
        tag=image_full,
        buildargs={
            "REGION": region_name
        },
        quiet=False,
        nocache=not cache,
        pull=pull
    )
    print("Built image {}".format(image_full))
    #repo = "{}.dkr.ecr.{}.amazonaws.com/"

    print("Pushing image {}".format(image_full))
    client.images.push(repo_uri, tag_tag)
    print("Pushed image {}".format(image_full))
    return image_full
    # 683880991063.dkr.ecr.us-east-1.amazonaws.com/columbo:inference


def image_exists(ecr, registry, repository, tag):
    return get_image(ecr, registry, repository, tag) is not None


def get_image(ecr, registry, repository, tag):
    try:
        img = ecr.batch_get_image(
            registryId=registry,
            repositoryName=repository,
            imageIds=[
                {
                    'imageTag': tag
                }
            ]
        )
    except Exception:
        # todo: check for RepositoryNotFoundException
        return None
    img = img['images']
    if len(img) > 0:
        return img[0]
    else:
        return None


def get_account(session):
    return session.client('sts').get_caller_identity().get('Account')


def parse_image(uri, account):
    """
    TRAINING_IMAGE = '683880991063.dkr.ecr.us-east-1.amazonaws.com/columbo-sagemaker-training:latest'
    """
    try:
        uri = uri.split("/")
        if len(uri) == 2:
            host, path = uri
            account = host.split(".")[0]
        elif len(uri) == 1:
            account = account
            path = uri[0]
        else:
            raise ValueError("Invalid uri")

        path = path.split(":")
        if len(path) == 2:
            name, tag = path
        elif len(path) == 1:
            name = path[0]
            tag = 'latest'
        else:
            raise "Invalid URI"
        return account, name, tag

    except Exception as e:
        raise ValueError("Invalid ECR URI: {}".format(uri)) from e


if __name__ == '__main__':
    """
    account = '683880991063'
    print(parse_image(
        '683880991063.dkr.ecr.us-east-1.amazonaws.com/columbo-sagemaker-training:latest', account=account))
    print(parse_image(
        '683880991063.dkr.ecr.us-east-1.amazonaws.com/columbo-sagemaker-training', account=account))
    print(parse_image('columbo-sagemaker-training:latest', account=account))
    print(parse_image('columbo-sagemaker-training', account=account))

    from boto3 import Session
    session = Session(profile_name='default')
    ecr = session.client('ecr')
    account = get_account(session)
    img = get_image(
        ecr, *parse_image('columbo-sagemaker-training', account=account))
    print(img)

    repo = ecr_describe_repository(ecr, account, 'columbo-sagemaker-training')
    print(repo)
    """
    from boto3 import Session
    session = Session(profile_name='default')
    ecr = session.client('ecr')
    account = get_account(session)
    for image in Images.ALL:
        img = ecr_ensure_image(
            path=image.path,
            tag=image.tag,
            accounts=image.accounts,
            session=session
        )
        print("Image: {}".format(img))



import json
from aws_sagemaker_remote.util.fields import get_field
import os
from urllib.parse import urlparse
import datetime

def processing_json(description):
    description['ProcessingInputs'] = {
        pi['InputName']: pi
        for pi in
        description.get('ProcessingInputs', {})
    }
    description['ProcessingOutputConfig'] = description.get(
        'ProcessingOutputConfig', {})
    description['ProcessingOutputConfig']['Outputs'] = {
        po['OutputName']: po
        for po in
        description['ProcessingOutputConfig'].get('Outputs', {})
    }
    return description

def json_read(path, field):
    with open(path, 'r') as f:
        data = json.load(f)
    data = processing_json(data)
    return get_field(data, field)


def urlparse_safe(url):
    try:
        return urlparse(url)
    except Exception:
        return None


def json_urlparse(url):
    if(url):
        uri = urlparse_safe(url)
        if uri and uri.scheme == 'json':
            path = [uri.netloc, uri.path]
            path = [p for p in path if p]
            path = "".join(path)
            if not uri.fragment:
                raise ValueError(
                    "URL fragment required for JSON inputs [{}]".format(url))
            url = json_read(path, uri.fragment)
            #print("Read value [{}] from [{}]: [{}]".format(
            #    uri.fragment, path, url))
    return url


def json_converter(o):
    if isinstance(o, datetime.datetime):
        return o.isoformat()  # __str__()
    else:
        raise ValueError("unknown: {}".format(o))


if __name__ == '__main__':
    print(json_urlparse(
        'json://output/dataprep.json#ProcessingOutputConfig.Outputs.output.S3Output.S3Uri'))



from .train import sagemaker_training_run
from .args import sagemaker_training_args, sagemaker_env_args
from argparse import ArgumentParser
import inspect
import os
import warnings
import types
from ..commands import Command, run_command


def sagemaker_training_handle(args, config, main, metrics=None):
    if args.sagemaker_run:
        # Remote processing
        sagemaker_training_run(
            args=args,
            config=config,
            metrics=metrics)
    else:
        # Local processing or on SageMaker container
        args = sagemaker_env_args(args=args, config=config)
        main(args)


class TrainingCommand(Command):
    def __init__(self, main, script=None, help=None, metrics=None, **training_args):
        super(TrainingCommand, self).__init__(help=help or 'Run training')
        if not script:
            script = main
        self.main = main
        self.script = script
        self.training_args = training_args
        self.config = None
        self.metrics = metrics

    def configure(self, parser: ArgumentParser):
        self.config = sagemaker_training_args(
            parser=parser,
            script=self.script,
            **self.training_args
        )

    def run(self, args):
        sagemaker_training_handle(
            args=args,
            config=self.config,
            main=self.main,
            metrics=self.metrics
        )


def sagemaker_training_main(
        main, script=None, script_fn=None, description=None, metrics=None,
        **training_args):
    r"""
    Entry point for training.

    Example
    -------

    .. code-block:: python

        from aws_sagemaker_remote import sagemaker_processing_main

        def main(args):
            # your code here
            pass

        if __name__ == '__main__':
            sagemaker_processing_main(
                main=main,
                # ... additional configuration
            )

    Parameters
    ----------
    main : function
        Main function. Must accept a single argument ``args`` (``argparse.Namespace``)
    script : str, optional
        Path to script file to execute. 
        Set to ``__file__`` for most use-cases.
        Empty or None defaults to file containing ``main``.
        Object interpreted as file containing the object.
    description: str, optional
        Script description for argparse
    metrics : dict, optional
        Metrics to record. Dictionary of metric name (str) to RegEx that extracts metric (str).
        See `SageMaker Training Metrics Docs <https://docs.aws.amazon.com/sagemaker/latest/dg/training-metrics.html#define-train-metric-regex>`_
    \**training_args : dict, optional
        Keyword arguments to :meth:`aws_sagemaker_remote.training.args.sagemaker_training_args`
    """
    # todo: detect module, class, etc. files
    command = TrainingCommand(
        main=main, script=script,
        metrics=metrics, **training_args
    )
    run_command(
        command=command, description=description
    )



from glob import glob
import os


def get_relative(file, base):
    if not file.startswith(base):
        raise ValueError("Expected [{}] to start with [{}]".format(
            file, base)
        )
    assert file[len(base)] in ["\\", "/"]
    return file[len(base)+1:]


def glob_relative(path, pattern):
    base = os.path.abspath(path)
    glob_path = os.path.join(base, "**", pattern)
    files = glob(glob_path, recursive=True)
    files = (os.path.abspath(file) for file in files)
    files = ((file, get_relative(file, base)) for file in files)
    return files



import multiprocessing
import traceback


def wrap_worker(fn, *args):
    semaphore.release()
    try:
        fn(*args)
    except Exception as e:
        print("exception in worker: {}".format(e))
        traceback.print_exc()
        raise e


def init_child(semaphore_):
    global semaphore
    semaphore = semaphore_


def run_workers(workers, fn, data, *args, expand=False, queue_size=100):
    assert workers > 0
    if workers > 1:
        assert queue_size > 0
        sem = multiprocessing.Semaphore(queue_size)
        with multiprocessing.Pool(workers, initializer=init_child, initargs=(sem,)) as pool:
            for datum in data:
                sem.acquire()
                if expand:
                    pool.apply_async(wrap_worker, (fn, *datum, *args))
                else:
                    pool.apply_async(wrap_worker, (fn, datum, *args))
            pool.close()
            pool.join()
    else:
        for datum in data:
            if expand:
                fn(*datum, *args)
            else:
                fn(datum, *args)



import sys
import json
from sagemaker.amazon.common import read_recordio


class PipeIterator(object):
    def __init__(self, path, size=1, count=0):
        self.path = path
        self.size = size
        self.count = count

    def __iter__(self):
        while True:
            with open("{}_{}".format(self.path, self.count), 'rb') as f:
                self.count += 1
                for rec in chunk_iterable(read_recordio(f), self.size, last='error'):
                    yield rec


def pipe_iterator(path, size=1):
    count = 0
    while True:
        with open("{}_{}".format(path, count), 'rb') as f:
            for rec in chunk_iterable(read_recordio(f), size, last='error'):
                yield rec
        count += 1


def read_jsonlines(path):
    with open(path) as f:
        for l in f:
            l = l.strip()
            if l:
                yield json.loads(l)


def chunk_iterable(it, size, last='skip'):
    assert last in ['skip', 'yield', 'error']
    i = iter(it)
    ret = []
    try:
        if last in ['yield', 'error']:
            while True:
                for _ in range(size):
                    ret.append(next(i))
                yield ret
                ret = []
        else:
            while True:
                yield [next(i) for _ in range(size)]
    except StopIteration:
        if last == 'yield':
            yield ret
        elif last == 'error':
            if len(ret):
                raise ValueError(
                    "Incomplete final chunk in chunk_iterable (len: {}, contents: {})".format(len(ret), [len(r) for r in ret]))
    finally:
        del i


if __name__ == '__main__':
    it = range(100)
    print(list(chunk_iterable(it, 7, last='skip')))
    print(list(chunk_iterable(it, 7, last='yield')))
    print(list(chunk_iterable(it, 10, last='error')))
    try:
        print(list(chunk_iterable(it, 7, last='error')))
    except ValueError as e:
        print(e)



from setuptools import find_packages, setup
import os
with open(os.path.abspath(os.path.join(__file__, '../README.rst')), encoding='utf-8') as f:
    long_description = f.read()
setup(name='aws-sagemaker-remote',
      version='0.0.50',
      author='Ben Striner',
      author_email="bstriner@gmail.com",
      url='https://github.com/bstriner/aws-sagemaker-remote',
      install_requires=[
          'sagemaker',
          'sagemaker-experiments',
          'click',
          'docker'
      ],
      package_data={'': [
          '**/Dockerfile',
          '**/*.sh',
          '**/*.txt',
          '**/*.yaml',
          '**/*.bat'
      ]},
      include_package_data=True,
      entry_points={
          'console_scripts': [
              'aws-sagemaker-remote=aws_sagemaker_remote.cli:cli'
          ]
      },
      description="Simplify running processing and training remotely on AWS SageMaker",
      packages=find_packages(),
      long_description=long_description,
      long_description_content_type='text/x-rst'
      )
